#pragma once
#include "simpleCL.h"	// The SimpleOpenCL library
#include <iostream>		// For input/output
#include <stdlib.h>		// For random function
#include <time.h>		// For random function initialization
#include <string>		// For string functions
#include "Utils.h"
using namespace std;

const bool printFullArray = true;		// Indicate if the full output arrays have to be printed
size_t maxWorkGroupSize;


Result executeAdvancedRecursive(int inputSize, int *input, int depth) {
	debug("\n");
	// Check if we need to calculate on one or more workgroups
	if (inputSize <= (maxWorkGroupSize*2)) {
		// Calculate on 1 workgroup
		unsigned int height = ceil(log2f(inputSize)) + 1;
		int paddedInputSize = pow(2, (height - 1));
		size_t bytesGPU = paddedInputSize*sizeof(int);
		Result result;

		// Arrays
		int *inputGPU = padInputTo(input, inputSize, paddedInputSize);
		int *tempGPU = new int[paddedInputSize*height];
		int *outputGPU = new int[paddedInputSize];

		debug("%d: Calculating Prefix Sum for %d values\n", depth, inputSize);

		// Get correct kernel
		sclSoft software = sclGetCLSoftware("GPUSimple2D.cl", "prefixSum", getHardware());

		// Setup global and local sizes
		size_t globalSize[2], localSize[2];
		globalSize[0] = paddedInputSize / 2;
		globalSize[1] = 1;
		localSize[0] = paddedInputSize / 2;
		localSize[1] = 1;
		debug("%d:  localSize=[%d, %d]\n", depth, localSize[0], localSize[1]);
		debug("%d:  globalSize=[%d, %d]\n", depth, globalSize[0], globalSize[1]);
		// Launch kernel with correct arguments
		cl_event kernelResult = NULL;
		if (printFullArray) {   // Only when debugging the temporary array will be accesible from the host
			kernelResult = sclManageArgsLaunchKernel(getHardware(), software, globalSize, localSize,
				" %r %w %R %a %a ",
				bytesGPU, inputGPU,		// Input array
				bytesGPU, outputGPU,	// Output array
				bytesGPU * height, tempGPU,		// Temporary array for the tree
				sizeof(unsigned int), (void*)&paddedInputSize,
				sizeof(unsigned int), (void*)&height
				);
		}
		else {
			kernelResult = sclManageArgsLaunchKernel(getHardware(), software, globalSize, localSize,
				" %r %w %g %a %a ",
				bytesGPU, inputGPU,		// Input array
				bytesGPU, outputGPU,	// Output array
				bytesGPU * height,				// Temporary array for the tree
				sizeof(unsigned int), (void*)&paddedInputSize,
				sizeof(unsigned int), (void*)&height
				);
		}
		// Print temporary array
		if (printFullArray) {
			printMultiArrayDepth(depth, paddedInputSize, height, tempGPU);
		}

		// Print the time it took to run the program
		cl_ulong time = sclGetEventTime(getHardware(), kernelResult);
		result.executionTime = time;
		debug("%d: Execution time: %lu nanoseconds, %0.3f milliseconds\n", depth, time, time / 1000000.0);
		result.output = outputGPU;

		// Release host memory
		delete[] inputGPU;
		//delete[] input;
		delete[] tempGPU;
		return result;
	}
	else {
		// Calculate using multiple workgroups
		int paddedInputSize = pow(2, ceil(log2f(inputSize)));	// Input padded to power of 2
		int blocks = ceil(((double)paddedInputSize) / 2 / maxWorkGroupSize);
		int width = maxWorkGroupSize * 2;
		int realInputSize = blocks * width;
		int height = ceil(log2f(width)) + 1;
		debug("%d: realInputSize=%d, width=%d, blocks=%d, height=%d\n", depth, realInputSize, width, blocks, height);
		debug("%d: Input size: %d, paddedInputSize: %d, realInputSize: %d\n", depth, inputSize, paddedInputSize, realInputSize);

		// Arrays and variables
		int *inputGPU = padInputTo(input, inputSize, realInputSize);
		int *tempGPU = new int[realInputSize*height];
		int *outputGPU = new int[realInputSize];
		int *sums = new int[blocks];
		size_t sumsBytes = blocks*sizeof(int);
		size_t bytesGPU = realInputSize*sizeof(int);			// Bytes the input/output needs
		Result result;											// Result of the kernel execution

		// Get correct kernels
		sclSoft prefixSum = sclGetCLSoftware("GPUAdvanced.cl", "prefixSum", getHardware());
		sclSoft adder = sclGetCLSoftware("GPUAdvanced.cl", "adder", getHardware());

		// Setup global and local sizes
		size_t globalSize[2], localSize[2];
		globalSize[0] = realInputSize / 2;
		globalSize[1] = 1;
		localSize[0] = width / 2;
		localSize[1] = 1;
		debug("%d:  localSize=[%d, %d]\n", depth, localSize[0], localSize[1]);
		debug("%d:  globalSize=[%d, %d]\n", depth, globalSize[0], globalSize[1]);
		unsigned int workSizeUnsigned = width;

		// Launch prefixsum kernel for partial prefix sums
		debug("%d: bytesGPU=%d, sumsBytes=%d, height=%d\n", depth, bytesGPU, sumsBytes, height);
		cl_event kernelResult1 = NULL;
		char *arguments = "input%r temp%g output%w sums%w inputsize%a height%a blockwidth%a";
		if (printFullArray) {
			arguments = "input%r temp%w output%w sums%w inputsize%a height%a blockwidth%a";
			kernelResult1 = sclManageArgsLaunchKernel(getHardware(), prefixSum, globalSize, localSize,
				arguments,
				bytesGPU, inputGPU,				// Input array
				bytesGPU * height, tempGPU,		// Temporary array for the trees (one tree per workgroup)
				bytesGPU, outputGPU,
				sumsBytes, sums,
				sizeof(unsigned int), (void*)&realInputSize,
				sizeof(unsigned int), (void*)&height,
				sizeof(unsigned int), (void*)&workSizeUnsigned
				);
		}
		else {
			kernelResult1 = sclManageArgsLaunchKernel(getHardware(), prefixSum, globalSize, localSize,
				arguments,
				bytesGPU, inputGPU,				// Input array
				bytesGPU * height,		// Temporary array for the trees (one tree per workgroup)
				bytesGPU, outputGPU,
				sumsBytes, sums,
				sizeof(unsigned int), (void*)&realInputSize,
				sizeof(unsigned int), (void*)&height,
				sizeof(unsigned int), (void*)&workSizeUnsigned
				);
		}
		debug("%d: Running scan kernel complete\n", depth);

		if (printFullArray) {
			// Print temporary array
			printMultiArrayDepth(depth, realInputSize, height, tempGPU);
			debug("%d: Scan output: ", depth);
		}

		// Calculate sums for sum array and print them
		if (printFullArray) {
			debug("%d: Raw sums array: ", depth);
			printArray(blocks, sums);
		}
		Result sumsResult = executeAdvancedRecursive(blocks, sums, depth+1);
		if (printFullArray) {
			debug("\n%d: Scanned sum array: ", depth);
			printArray(blocks, sumsResult.output);
		}

		// Add sums to the output
		cl_event kernelResult2 = NULL;
		kernelResult2 = sclManageArgsLaunchKernel(getHardware(), adder, globalSize, localSize,
			"data%R sums%r blockwidth%a",
			bytesGPU, outputGPU,				// Data array
			sumsBytes, sumsResult.output,
			sizeof(unsigned int), (void*)&workSizeUnsigned
			);
		debug("\n%d: Running adder kernel complete\n", depth);
		delete[] sumsResult.output;
		delete[] sums;

		// Print the time it took to run the program
		cl_ulong timeSum = sclGetEventTime(getHardware(), kernelResult1);
		cl_ulong timeAdder = sclGetEventTime(getHardware(), kernelResult1);
		debug("%d: Execution times:\n", depth);
		debug("%d:  Initial sum: %lu nanoseconds, %0.3f milliseconds\n", depth, timeSum, timeSum / 1000000.0);
		debug("%d:  Sums scan:   %lu nanoseconds, %0.3f milliseconds\n", depth, sumsResult.executionTime, sumsResult.executionTime / 1000000.0);
		debug("%d:  Adder:       %lu nanoseconds, %0.3f milliseconds\n", depth, timeAdder, timeAdder / 1000000.0);

		result.executionTime = timeSum + sumsResult.executionTime + timeAdder;
		result.output = outputGPU;

		// Release host memory
		delete[] inputGPU;
		delete[] tempGPU;
		sclReleaseClSoft(prefixSum);
		sclReleaseClSoft(adder);
		return result;
	}
}

Result executeAdvanced(int inputSize, int *input) {
	// General initialization
	maxWorkGroupSize = getMaximumWorkgroupSize(getHardware());
	//maxWorkGroupSize = 2; // TODO remove (debug mode)
	debug("0: Using maximum workgroup size of %d.\n", maxWorkGroupSize);

	// Print hardware info
	char* value = getHardwareName(getHardware());
	debug("0: Using hardware device: %s\n", value);
	Result scanResult = executeAdvancedRecursive(inputSize, input, 1);

	// Calculate the result with the CPU
	debug("\n0: GPU calculation complete, CPU calculation started.\n");
	int *outputCPU = calculateCPU(inputSize, input);

	// Display complete CPU and GPU arrays
	if (printFullArray) {
		printArraysDepth(0, inputSize, input, outputCPU, scanResult.output);
	}

	// Compare the results of the CPU with the GPU
	int incorrect = 0;
	for (int i = 0; i < inputSize; i++) {
		if (outputCPU[i] != scanResult.output[i]) {
			incorrect++;
		}
	}
	debug("0:\n");
	if (incorrect == 0) {
		debug("0: Correct result, all %d values match.\n", inputSize);
	}
	else {
		debug("0: FAILED, %d incorrect results out of %d output values.\n", incorrect, inputSize);
	}

	debug("0: Total execution time: %lu nanoseconds, %0.3f milliseconds\n", scanResult.executionTime, scanResult.executionTime / 1000000.0);

	delete[] outputCPU;
	return scanResult;
}


